/**
 * Deque
 * safe against Thread InterruptedException
 * 
 */

//package edu.unima.ky.parallel;

import java.io.*;

public class Deque  implements Serializable {
  /**
   * object storage
   */
  private Object[] buffer;

  /**
   * buffer size
   */
  private int size;

  /**
   * index to the first filled cell
   */
  private int front;

  /**
   * index to the first empty cell
   */
  private int rear;

  /**
   * semaphore indicating an empty cell
   */
  private Semaphore empty;

  /**
   * semaphore indicating a filled cell
   */
  private Semaphore full;

  /**
   * mutex for send
   */
  private Object snd;

  /**
   * mutex for receive
   */
  private Object rcv;

  /**
   * undefined value
   */
  public final static String undefined = "undefined value";



  /**
   * Constructs a Deque
   */
  public Deque(int init) {
    buffer = new Object[init];
    size = init; front = 0; rear = 0;
    empty = new Semaphore(init);
    full = new Semaphore(0);
    snd = new Object();
    rcv = new Object();
  }


  /**
   * Put an object to the deque
   */
  public void put(Object v) throws InterruptedException {
    empty.P(); // return without Object send if Interrupted
    synchronized (snd) {
      if (rear < 0) rear += size;
      if (rear >= size) rear -= size;
      buffer[rear] = v;
      rear++;
    }
    full.V();
  }


  /**
   * Push an object to the deque
   */
  public void push(Object v) throws InterruptedException {
    put(v);
  }


  /**
   * Get an object from the deque
   */
  public Object get() throws InterruptedException {
    Object v = null;
    full.P();  // return without Object if Interrupted
    synchronized (rcv) {
      if (front >= size) front -= size;
      v = buffer[front];
      front++;
    }
    empty.V();
    return v;
  }


  /**
   * Get an object from the deque within given timeout
   */
  public Object get(int m) throws InterruptedException {
    Object v = null;
    if ( !full.P(m) ) return undefined; 
        // return without Object if Interrupted
    synchronized (rcv) {
      if (front >= size) front -= size;
      v = buffer[front];
      front++;
    }
    empty.V();
    return v;
  }


  /**
   * Pop an object from the deque
   */
  public Object pop() throws InterruptedException {
    Object v = null;
    full.P();  // return without Object if Interrupted
    synchronized (snd) {
      rear--;
      if (rear < 0) rear += size;
      if (rear >= size) rear -= size;
      v = buffer[rear];
    }
    empty.V();
    return v;
  }


  /**
   * Pop an object from the deque within given timeout
   */
  public Object pop(int m) throws InterruptedException {
    Object v = null;
    if ( !full.P(m) ) return undefined; 
        // return without Object if Interrupted
    synchronized (snd) {
      rear--;
      if (rear < 0) rear += size;
      if (rear >= size) rear -= size;
      v = buffer[rear];
    }
    empty.V();
    return v;
  }


  /**
   * Test if deque is empty
   */
  public boolean empty() {
    return ( !full.isPositive() );
  }

}
